from typing import List

from adhesive.steps.AdhesiveTask import AdhesiveTask
from adhesive.graph.Workflow import Workflow


class AdhesiveProcess:
    """
    An Adhesive process. Holds the linkage between
    the graph, and the steps.
    """
    def __init__(self, id: str) -> None:
        self.steps: List[AdhesiveTask] = []
        self.workflow: Workflow = Workflow(id)



import asyncio
from typing import Callable, TypeVar, Optional

from adhesive.model.AdhesiveProcess import AdhesiveProcess
from adhesive.model.ConsoleUserTaskProvider import ConsoleUserTaskProvider
from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.steps.AdhesiveTask import AdhesiveTask
from adhesive.xml.bpmn import read_bpmn_file

T = TypeVar('T')
process = AdhesiveProcess('_root')


def task(*task_names: str) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        process.steps.append(AdhesiveTask(f, *task_names))
        return f

    return wrapper_builder


# FIXME: they seem different types. Create base, and different implementations?
# Doesn't seem right using the regular AdhesiveTask here.
def usertask(*task_names: str) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        process.steps.append(AdhesiveTask(f, *task_names))
        return f

    return wrapper_builder


def bpmn_build(file_name: str,
               ut_provider: Optional['UserTaskProvider'] = None,
               wait_tasks: bool=True):
    """ Start a build that was described in BPMN """
    process.workflow = read_bpmn_file(file_name)

    if ut_provider is None:
        ut_provider = ConsoleUserTaskProvider()

    fn = WorkflowExecutor(
        process,
        ut_provider=ut_provider,
        wait_tasks=wait_tasks).execute()

    asyncio.get_event_loop().run_until_complete(fn)


def bpmn_build_async(
        file_name: str,
        ut_provider: Optional['UserTaskProvider'] = None,
        wait_tasks: bool=True):
    """ Start a build that was described in BPMN """
    process.workflow = read_bpmn_file(file_name)

    if ut_provider is None:
        ut_provider = ConsoleUserTaskProvider()

    return WorkflowExecutor(
        process,
        ut_provider=ut_provider,
        wait_tasks=wait_tasks).execute()


from adhesive.model.UserTaskProvider import UserTaskProvider



from typing import List, cast

from adhesive.graph.Edge import Edge
from adhesive.graph.BaseTask import BaseTask
from adhesive.graph.Workflow import Workflow
from adhesive.model.ActiveEvent import ActiveEvent


class Gateway(BaseTask):
    pass


class NonWaitingGateway(Gateway):
    pass


class WaitingGateway(Gateway):
    pass



from enum import Enum
from typing import Any, Dict, Optional, Callable, Union
import uuid


class ActiveEventState(Enum):
    NEW = 'NEW'
    PROCESSING = 'PROCESSING'
    WAITING = 'WAITING'
    ERROR = 'ERROR'
    RUNNING = 'RUNNING'
    ROUTING = 'ROUTING'
    DONE_CHECK = 'DONE_CHECK'
    DONE_END_TASK = 'DONE_END_TASK'
    DONE = 'DONE'


STATE_INDEX = {
    'NEW': 0,
    'PROCESSING': 1,
    'WAITING': 2,
    'ERROR': 3,
    'RUNNING': 4,
    'ROUTING': 5,
    'DONE_CHECK': 6,
    'DONE_END_TASK': 7,
    'DONE': 8,
}


class ActiveEventStateChangeEvent(object):
    """
    Event that gets on all the before/after callbacks that are
    triggered on state changes.
    """

    def __init__(self,
                 previous_state: Optional[ActiveEventState],
                 target_state: ActiveEventState,
                 data: Any) -> None:
        """
        Create a new event.

        :param ActiveEventState previous_state: The state that the state machine is transitioning from.
        :param ActiveEventState target_state: The state that the state machine is transitioning to.
        :param object data: Optional data that is passed in the event.
        """
        self._previous_state = previous_state
        self._target_state = target_state
        self.data = data
        self._cancelled = False

    def cancel(self) -> None:
        """
        Cancel the current transition.
        """
        self._cancelled = True

    @property
    def cancelled(self) -> bool:
        """
        Is the current transition cancelled.
        :return:
        """
        return self._cancelled

    @property
    def previous_state(self) -> Optional[ActiveEventState]:
        """
        The state from which we're transitioning.
        :return:
        """
        return self._previous_state

    @property
    def target_state(self) -> ActiveEventState:
        """
        Thestate towards we're transitioning.
        :return:
        """
        return self._target_state


class ActiveEventStateException(Exception):
    pass


transition_set: Dict[int, bool] = dict()
link_map: Dict[ActiveEventState, Dict[str, ActiveEventState]] = dict()


def register_transition(name: Optional[str], from_state: ActiveEventState, to_state: ActiveEventState) -> None:
    transition_set[STATE_INDEX[from_state.value] << 14 | STATE_INDEX[to_state.value]] = True

    if not name:
        return

    fromMap = link_map.get(from_state.value)

    if not fromMap:
        fromMap = link_map[from_state.value] = dict()

    fromMap[name] = to_state


register_transition('process', ActiveEventState.NEW, ActiveEventState.PROCESSING)
register_transition('error', ActiveEventState.NEW, ActiveEventState.DONE)
register_transition('route', ActiveEventState.PROCESSING, ActiveEventState.ROUTING)
register_transition('wait_check', ActiveEventState.PROCESSING, ActiveEventState.WAITING)
register_transition('run', ActiveEventState.PROCESSING, ActiveEventState.RUNNING)
register_transition('done', ActiveEventState.PROCESSING, ActiveEventState.DONE)
register_transition('run', ActiveEventState.WAITING, ActiveEventState.RUNNING)
register_transition('done', ActiveEventState.WAITING, ActiveEventState.DONE)
register_transition('error', ActiveEventState.WAITING, ActiveEventState.ERROR)
register_transition('route', ActiveEventState.RUNNING, ActiveEventState.ROUTING)
register_transition('error', ActiveEventState.RUNNING, ActiveEventState.ERROR)
register_transition('route', ActiveEventState.ERROR, ActiveEventState.ROUTING)
register_transition('done_check', ActiveEventState.ERROR, ActiveEventState.DONE_CHECK)
register_transition('done_check', ActiveEventState.ROUTING, ActiveEventState.DONE_CHECK)
register_transition('done', ActiveEventState.DONE_CHECK, ActiveEventState.DONE)
register_transition('done_end_task', ActiveEventState.DONE_CHECK, ActiveEventState.DONE_END_TASK)
register_transition('done', ActiveEventState.DONE_END_TASK, ActiveEventState.DONE)


ChangeStateEventListener = Union[
    Callable[[], Optional[ActiveEventState]],
    Callable[[ActiveEventStateChangeEvent], Optional[ActiveEventState]]
]


class ActiveEventStateMachine(object):
    def __init__(self, initial_state: Optional[ActiveEventState]=None) -> None:
        self._transition_listeners: Dict[str, EventListener] = dict()
        self._data_listeners: Dict[str, EventListener] = dict()
        self._initial_state = initial_state or ActiveEventState.NEW

        self._transition_listeners['NEW'] = EventListener()
        self._transition_listeners['PROCESSING'] = EventListener()
        self._transition_listeners['WAITING'] = EventListener()
        self._transition_listeners['ERROR'] = EventListener()
        self._transition_listeners['RUNNING'] = EventListener()
        self._transition_listeners['ROUTING'] = EventListener()
        self._transition_listeners['DONE_CHECK'] = EventListener()
        self._transition_listeners['DONE_END_TASK'] = EventListener()
        self._transition_listeners['DONE'] = EventListener()
        self._data_listeners['NEW'] = EventListener()
        self._data_listeners['PROCESSING'] = EventListener()
        self._data_listeners['WAITING'] = EventListener()
        self._data_listeners['ERROR'] = EventListener()
        self._data_listeners['RUNNING'] = EventListener()
        self._data_listeners['ROUTING'] = EventListener()
        self._data_listeners['DONE_CHECK'] = EventListener()
        self._data_listeners['DONE_END_TASK'] = EventListener()
        self._data_listeners['DONE'] = EventListener()
        self._currentState = None  # type: Optional[ActiveEventState]
        self._current_change_state_event = None  # type: Optional[ActiveEventStateChangeEvent]

    @property
    def state(self) -> ActiveEventState:
        self._ensure_state_machine_initialized()
        assert self._currentState

        return self._currentState

    def process(self, data: Any=None) -> ActiveEventState:
        return self.transition("process", data)

    def error(self, data: Any=None) -> ActiveEventState:
        return self.transition("error", data)

    def route(self, data: Any=None) -> ActiveEventState:
        return self.transition("route", data)

    def wait_check(self, data: Any=None) -> ActiveEventState:
        return self.transition("wait_check", data)

    def run(self, data: Any=None) -> ActiveEventState:
        return self.transition("run", data)

    def done(self, data: Any=None) -> ActiveEventState:
        return self.transition("done", data)

    def done_check(self, data: Any=None) -> ActiveEventState:
        return self.transition("done_check", data)

    def done_end_task(self, data: Any=None) -> ActiveEventState:
        return self.transition("done_end_task", data)

    def _ensure_state_machine_initialized(self) -> None:
        if not self._currentState:
            self._change_state_impl(self._initial_state, None)

    def changeState(self, targetState: ActiveEventState, data: Any=None) -> ActiveEventState:
        self._ensure_state_machine_initialized()
        return self._change_state_impl(targetState, data)

    def _change_state_impl(self, targetState: ActiveEventState, data: Any=None) -> ActiveEventState:
        if not targetState:
            raise Exception("No target state specified. Can not change the state.")

        # this also ignores the fact that maybe there is no transition
        # into the same state.
        if targetState == self._currentState:
            return targetState

        state_change_event: ActiveEventStateChangeEvent = ActiveEventStateChangeEvent(self._currentState, targetState, data)

        if self._currentState and \
                not transition_set.get(STATE_INDEX[self._currentState.value] << 14 | STATE_INDEX[targetState.value]):
            print("No transition exists between %s -> %s." % (self._currentState.value, targetState.value))
            return self._currentState

        if self._current_change_state_event:
            # The previous_state if it's None, is only set when the initial transition happens into
            # the start state. Then only the *AFTER* callbacks are being invoked, not the *BEFORE*,
            # because entering the initial state is not cancellable. The state machine is assumed to
            # be in the initial state.
            #
            # Because of that, if there is another _current_change_state_event (i.e. set on a *BEFORE*
            # callback), it's after being already in the initial state, hence having a previous_state.
            assert self._currentState
            assert self._current_change_state_event.previous_state

            raise ActiveEventStateException(
                "The ActiveEventStateMachine is already in a changeState (%s -> %s). "
                "Transitioning the state machine (%s -> %s) in `before` events is not supported." % (
                    self._current_change_state_event.previous_state.value,
                    self._current_change_state_event.target_state.value,
                    self._currentState.value,
                    targetState.value
                ))

        self._current_change_state_event = state_change_event

        if state_change_event.previous_state:
            self._transition_listeners[state_change_event.previous_state.value]\
                .fire(EventType.BEFORE_LEAVE, state_change_event)

        self._transition_listeners[state_change_event.target_state.value]\
            .fire(EventType.BEFORE_ENTER, state_change_event)

        # The event can't be cancelled in the initial state.
        if state_change_event.cancelled:
            assert self._currentState
            return self._currentState

        self._currentState = targetState
        self._current_change_state_event = None

        if state_change_event.previous_state:
            self._transition_listeners[state_change_event.previous_state.value]\
                .fire(EventType.AFTER_LEAVE, state_change_event)

        self._transition_listeners[state_change_event.target_state.value]\
            .fire(EventType.AFTER_ENTER, state_change_event)

        return self._currentState

    def transition(self, link_name: str, data: Any=None) -> ActiveEventState:
        """
        Transition into another state following a named transition.

        :param str link_name:
        :param object data:
        :return: ActiveEventState
        """
        self._ensure_state_machine_initialized()

        assert self._currentState

        source_state = link_map.get(self._currentState.value)

        if not source_state:
            return self._currentState

        if link_name not in source_state:
            print("There is no transition named `%s` starting from `%s`." %
                  (link_name, self._currentState.value))

            return self._currentState

        targetState = source_state[link_name]

        if not targetState:
            return self._currentState

        return self.changeState(targetState, data)

    def before_enter(self, state: ActiveEventState, callback: ChangeStateEventListener):
        """
        Add a transition listener that will fire before entering a new state.
        The transition can still be cancelled at this stage via `ev.cancel()`
        in the callback.

        :param ActiveEventState state:
        :param Function callback:
        :return:
        """
        return self._transition_listeners[state.value].add_listener(EventType.BEFORE_ENTER, callback)

    def after_enter(self, state: ActiveEventState, callback: ChangeStateEventListener):
        """
        Add a transition listener that will fire after the new state is entered.
        The transition can not be cancelled at this stage.
        :param ActiveEventState state:
        :param callback:
        :return:
        """
        return self._transition_listeners[state.value].add_listener(EventType.AFTER_ENTER, callback)

    def before_leave(self, state: ActiveEventState, callback: ChangeStateEventListener):
        """
        Add a transition listener that will fire before leaving a state.
        The transition can be cancelled at this stage via `ev.cancel()`.

        :param ActiveEventState state:
        :param callback:
        :return:
        """
        return self._transition_listeners[state.value].add_listener(EventType.BEFORE_LEAVE, callback)

    def after_leave(self, state: ActiveEventState, callback: ChangeStateEventListener):
        """
        Add a transition listener that will fire after leaving a state.
        The transition can not be cancelled at this stage.

        :param ActiveEventState state:
        :param callback:
        :return:
        """
        return self._transition_listeners[state.value].add_listener(EventType.AFTER_LEAVE, callback)

    def on_data(self, state: ActiveEventState, callback: Callable[[Any], Optional[ActiveEventState]]):
        """
        Add a data listener that will be called when data is being pushed for that transition.

        :param ActiveEventState state:
        :param callback:
        :return:
        """
        return self._data_listeners[state.value].add_listener(EventType.DATA, callback)

    def forward_data(self, new_state: ActiveEventState, data: Any) -> None:
        """
        Changes the state machine into the new state, then sends the data
        ignoring the result. This is so on `onData` calls we can just
        short-circuit the execution using: `return stateMachine.forwardData(..)`

        @param new_state The state to transition into.
        @param data The data to send.
        """
        self.send_data(new_state, data)

        return None

    def send_state_data(self, new_state: ActiveEventState, data: Any) -> ActiveEventState:
        """
        Sends the data into the state machine, to be processed by listeners
        registered with `onData`.
        @param new_state
        @param data The data to send.
        """
        self._ensure_state_machine_initialized()

        assert self._currentState

        self.changeState(new_state, data)

        target_state = self._data_listeners[self._currentState.value].fire(EventType.DATA, data)

        if target_state:
            return self.changeState(target_state, data)

        return self._currentState

    def send_data(self,
                  data: Any=None,
                  state: Optional[ActiveEventState]=None) -> ActiveEventState:
        """
        Transitions first the state machine into the new state, then it
        will send the data into the state machine.
        @param newState
        @param data
        """
        self._ensure_state_machine_initialized()

        assert self._currentState

        if state:
            self.changeState(state)

        target_state = self._data_listeners[self._currentState.value]\
            .fire(EventType.DATA, data)

        if target_state:
            return self.changeState(target_state, data)

        return self._currentState


class EventType(Enum):
    BEFORE_ENTER = 'before-enter'
    BEFORE_LEAVE = 'before-leave'
    AFTER_LEAVE = 'after-leave'
    AFTER_ENTER = 'after-enter'
    DATA = 'data'


class EventListenerRegistration(object):
    def __init__(self, event_listener, callback_id):
        self._event_listener = event_listener
        self._callback_id = callback_id

    def detach(self):
        self._event_listener.eventListeners.pop(self._callback_id)


class EventListener(object):
    def __init__(self):
        self.registered = dict()

    def add_listener(self, event_name, callback):
        event_listeners = self.registered.get(event_name.value)

        if not event_listeners:
            event_listeners = self.registered[event_name.value] = dict()

        callback_id = uuid.uuid4()
        event_listeners[callback_id] = callback

        return EventListenerRegistration(self, callback_id)

    def fire(self, event_type, ev):
        result = None

        if not self.registered.get(event_type.value):
            return

        listeners = self.registered[event_type.value]

        for callback in listeners.values():
            potential_result = callback.__call__(ev)

            if potential_result and result:
                raise ActiveEventStateException("Data is already returned")

            result = potential_result

        return result



from typing import List

from adhesive.graph.Edge import Edge
from adhesive.graph.Gateway import Gateway
from adhesive.graph.BaseTask import BaseTask
from adhesive.graph.Workflow import Workflow
from adhesive.model.ActiveEvent import ActiveEvent


class GatewayController:
    @staticmethod
    def route_single_output(
            workflow: Workflow,
            gateway: Gateway,
            event: ActiveEvent) -> List[Edge]:

        default_edge = None
        result_edge = None

        edges = workflow.get_outgoing_edges(gateway.id)

        for edge in edges:
            if not edge.condition:
                if default_edge is not None:
                    raise Exception(f"Duplicate default edge for gateway {gateway.id}.")

                default_edge = edge
                continue

            if eval(edge.condition, globals(), event.context.as_mapping()):
                if result_edge is not None:
                    raise Exception(f"Duplicate output edge for gateway {gateway.id}")

                result_edge = edge
                continue

        if result_edge is None and default_edge is not None:
            result_edge = default_edge

        if not result_edge:
            raise Exception(f"No branch matches on gateway {gateway.id}")

        return [result_edge]

    @staticmethod
    def route_all_outputs(
            workflow: Workflow,
            task: BaseTask,
            event: ActiveEvent) -> List[Edge]:

        result_edges = []
        edges = workflow.get_outgoing_edges(task.id)

        for edge in edges:
            # if we have no condition on the edge, we create an event for it
            if edge.condition and not eval(edge.condition, globals(), event.context.as_mapping()):
                continue

            result_edges.append(edge)

        return result_edges




import unittest
from typing import Optional, List

from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.model.UserTaskProvider import UserTaskProvider
from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.steps.WorkflowContext import WorkflowContext
from adhesive.xml.bpmn import read_bpmn_file
from .test_tasks import adhesive, _async


class UIBuilder():
    def __init__(self,
                 context: WorkflowContext):
        self.context = context

    def add_text_input(self,
                       name: str,
                       title: Optional[str] = None,
                       value: Optional[str] = None) -> None:
        self.context.data[name] = name

    def add_combo_box(self,
                      name: str,
                      title: Optional[str] = None,
                      values: Optional[List[str]] = None):
        self.context.data[name] = values[0]

    def add_default_button(self,
                           name: str) -> None:
        self.context.data[name] = name


class TestUserTaskProvider(UserTaskProvider):
    def __init__(self):
        super(TestUserTaskProvider, self).__init__()

    def register_event(self,
                       executor: WorkflowExecutor,
                       event: ActiveEvent) -> None:
        try:
            ui = UIBuilder(event.context)

            adhesive_task = executor.tasks_impl[event.task.id]
            context = adhesive_task.invoke_user_task(event.context, ui)

            event.future.set_result(context)
        except Exception as e:
            event.future.set_exception(e)


class TestUserTaskBasic(unittest.TestCase):
    def test_link_back_execution(self):
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/user-task.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process,
                                             ut_provider=TestUserTaskProvider())
        data = _async(workflow_executor.execute())

        self.assertEqual("OK", data.OK)
        self.assertEqual("Cancel", data.Cancel)
        self.assertEqual("branch", data.branch)
        self.assertEqual("12.0", data.version)

        self.assertFalse(workflow_executor.events)


if __name__ == '__main__':
    unittest.main()



from typing import Dict, List, Iterator, Optional

import networkx as nx

from adhesive.graph.BoundaryEvent import BoundaryEvent, ErrorBoundaryEvent
from .BaseTask import BaseTask
from .StartEvent import StartEvent
from .EndEvent import EndEvent
from .Edge import Edge


class Workflow(BaseTask):
    """
    A workflow for the build
    """
    def __init__(self,
                 id: str,
                 name: str = '[root process]') -> None:
        super(Workflow, self).__init__(id, name)

        self._start_events: Dict[str, StartEvent] = dict()
        self._tasks: Dict[str, BaseTask] = dict()
        self._edges: Dict[str, Edge] = dict()
        self._end_events: Dict[str, EndEvent] = dict()

        self._graph = nx.MultiDiGraph()

    @property
    def start_tasks(self) -> Dict[str, StartEvent]:
        return self._start_events

    @property
    def tasks(self) -> Dict[str, BaseTask]:
        return self._tasks

    @property
    def edges(self) -> Dict[str, Edge]:
        return self._edges

    @property
    def end_events(self) -> Dict[str, EndEvent]:
        return self._end_events

    def add_task(self, task: BaseTask) -> None:
        """ Add a task into the graph. """
        self._tasks[task.id] = task
        self._graph.add_node(task.id)

    def add_boundary_event(self, boundary_event: BoundaryEvent) -> None:
        self._tasks[boundary_event.id] = boundary_event
        self._graph.add_node(boundary_event.id)
        self._graph.add_edge(boundary_event.attached_task_id, boundary_event.id)

        if isinstance(boundary_event, ErrorBoundaryEvent):
            self._tasks[boundary_event.attached_task_id].error_task = boundary_event

    def add_edge(self, edge: Edge) -> None:
        """Add an edge into the graph. """
        self._edges[edge.id] = edge
        self._graph.add_edge(
            edge.source_id,
            edge.target_id)

    def add_start_event(self, event: StartEvent) -> None:
        self._start_events[event.id] = event
        self._tasks[event.id] = event

    def add_end_event(self, event: EndEvent) -> None:
        self._end_events[event.id] = event
        self._tasks[event.id] = event

    def get_outgoing_edges(self, task_id: str) -> List[Edge]:
        """ Get the outgoing edges. """
        result: List[Edge] = []

        for edge_id, edge in self._edges.items():
            if edge.source_id == task_id:
                result.append(edge)

        return result

    def has_incoming_edges(self, task: BaseTask) -> bool:
        for edge_id, edge in self._edges.items():
            if edge.target_id == task.id:
                return True

        return False

    def has_outgoing_edges(self, task: BaseTask) -> bool:
        for edge_id, edge in self._edges.items():
            if edge.source_id == task.id:
                return True

        return False

    def are_predecessors(self,
                         task: BaseTask,
                         potential_predecessors: Iterator[BaseTask]) -> bool:
        predecessors = list(potential_predecessors)
        for potential_predecessor in predecessors:
            # FIXME: cross subprocess exceptions are handled as no predecessors
            try:
                if nx.algorithms.has_path(
                        self._graph,
                        potential_predecessor.id,
                        task.id):
                    return True
            except Exception:
                pass

        return False



from typing import Optional


class BaseTask:
    """
    Create a task. A task must have a name.
    Later this will be bounded to an actual implementation
    to be executed.
    """
    def __init__(self,
                 _id: str,
                 name: str) -> None:
        self.id = _id
        self.name = name
        self.error_task: Optional['ErrorBoundaryEvent'] = None

    def __str__(self) -> str:
        return f"Task({self.id}): {self.name}"


from adhesive.graph.BoundaryEvent import ErrorBoundaryEvent



import re
from concurrent.futures import Future
from typing import Set, Optional, Dict, TypeVar, cast, Any

from adhesive.graph.BoundaryEvent import BoundaryEvent
from adhesive.graph.Gateway import Gateway, NonWaitingGateway, WaitingGateway
from adhesive.graph.Task import Task
from adhesive.graph.UserTask import UserTask
from adhesive.model.ActiveEventStateMachine import ActiveEventState
from adhesive.model.GatewayController import GatewayController
from adhesive.model.WorkflowExecutorConfig import WorkflowExecutorConfig
from adhesive.steps.WorkflowData import WorkflowData

T = TypeVar('T')

import concurrent.futures

from adhesive.graph.EndEvent import EndEvent
from adhesive.graph.StartEvent import StartEvent
from adhesive.graph.SubProcess import SubProcess
from adhesive.graph.BaseTask import BaseTask
from adhesive.graph.Workflow import Workflow
from adhesive.steps.AdhesiveTask import AdhesiveTask
from adhesive.graph.ExclusiveGateway import ExclusiveGateway

from .ActiveEvent import ActiveEvent
from .AdhesiveProcess import AdhesiveProcess

DONE_STATES = {
    ActiveEventState.DONE_CHECK,
    ActiveEventState.DONE_END_TASK,
    ActiveEventState.DONE,
}

ACTIVE_STATES = {
    ActiveEventState.NEW,
    ActiveEventState.PROCESSING,
    ActiveEventState.WAITING,
    ActiveEventState.RUNNING,
    ActiveEventState.ERROR,
    ActiveEventState.ROUTING,
}


class WorkflowExecutor:
    """
    An executor of AdhesiveProcesses.
    """
    pool = concurrent.futures.ProcessPoolExecutor()

    def __init__(self,
                 process: AdhesiveProcess,
                 ut_provider: Optional['UserTaskProvider'] = None,
                 wait_tasks: bool = True) -> None:
        self.process = process
        self.tasks_impl: Dict[str, AdhesiveTask] = dict()

        # A dictionary of events that are currently active. This is just to find out
        # the parent of an event, since from it we can also derive the current parent
        # workflow.
        self.events: Dict[str, ActiveEvent] = dict()
        self.futures: Dict[Any, str] = dict()
        self.ut_provider = ut_provider

        self.config = WorkflowExecutorConfig(wait_tasks=wait_tasks)

    async def execute(self) -> WorkflowData:
        """
        Execute the current events. This will ensure new events are
        generating for forked events.
        """
        workflow = self.process.workflow
        self.tasks_impl: Dict[str, AdhesiveTask] = dict()

        self._validate_tasks(workflow)

        fake_event = ActiveEvent(parent_id=None, task=workflow)
        fake_event.id = None

        root_event = self.clone_event(fake_event, workflow)

        def raise_exception(_ev):
            raise _ev.data

        root_event.state.after_enter(ActiveEventState.ERROR, raise_exception)

        await self.execute_workflow()

        return root_event.context.data

    async def execute_workflow(self) -> None:
        """
        Process the events in a workflow until no more events are available.
        For example an event is the start of the workflow. The events are
        then creating futures (i.e. actual stuff that's being processed)
        that in turn might generate new events.
        :return:
        """
        while self.events:
            pending_events = filter(lambda e: e.state.state == ActiveEventState.NEW,
                                    self.events.values())

            # except for new processes, and processes that are running with futures
            # the states should transition automatically.
            processed_events = list(pending_events)
            for pending_event in processed_events:
                pending_event.state.process()

            done_futures, not_done_futures = concurrent.futures.wait(
                self.futures.keys(),
                return_when=concurrent.futures.FIRST_COMPLETED,
                timeout=5)

            for future in done_futures:
                event_id = self.futures[future]
                try:
                    context = future.result()
                    self.events[event_id].state.route(context)
                except Exception as e:
                    self.events[event_id].state.error(e)
                finally:
                    del self.futures[future]

    def register_event(self,
                       event: ActiveEvent) -> ActiveEvent:
        """
        Register the event into a big map for finding, so we can access it later
        for parents and what not, without serializing the event graph to
        the subprocesses.

        :param event:
        :return:
        """
        if event.id in self.events:
            raise Exception(f"Event {event.id} is already registered as "
                            f"{self.events[event.id]}. Got a new request to register "
                            f"it as {event}.")

        print(f"Register {event}")

        self.events[event.id] = event

        return event

    def unregister_event(self,
                         event: ActiveEvent) -> None:
        if event.id not in self.events:
            raise Exception(f"{event} not found in events. Either the event was "
                            f"already terminated, either it was not registered.")

        print(f"Unregister {event}")

        del self.events[event.id]

    def get_parent(self,
                   event_id: str) -> ActiveEvent:
        """
        Find the parent of an event by looking at the event ids.
        :param event_id:
        :return:
        """
        event = self.events[event_id]
        parent = self.events[event.parent_id]

        return parent

    def _validate_tasks(self,
                        workflow: Workflow) -> None:
        """
        Recursively traverse the graph, and print to the user if it needs to implement
        some tasks.

        :param workflow:
        :return:
        """
        unmatched_tasks: Set[BaseTask] = set()

        for task_id, task in workflow.tasks.items():
            if isinstance(task, SubProcess):
                self._validate_tasks(task)
                continue

            # gateways don't have associated tasks with them.
            if isinstance(task, StartEvent) or \
                    isinstance(task, EndEvent) or \
                    isinstance(task, Gateway) or \
                    isinstance(task, BoundaryEvent):
                continue

            adhesive_step = self._match_task(task)

            self.tasks_impl[task_id] = adhesive_step

            if not adhesive_step:
                unmatched_tasks.add(task)

        if unmatched_tasks:
            print("Missing tasks implementations. Generate with:\n")
            for unmatched_task in unmatched_tasks:
                print(f"@adhesive.task('{re.escape(unmatched_task.name)}')")
                print("def task_impl(context):")
                print("    pass\n\n")

            raise Exception("Missing tasks implementations")

    def _match_task(self, task: BaseTask) -> Optional[AdhesiveTask]:
        for step in self.process.steps:
            if step.matches(task.name) is not None:
                return step

        return None

    def clone_event(self,
                    old_event: ActiveEvent,
                    task: BaseTask,
                    parent_id: Optional[str] = None) -> ActiveEvent:

        if parent_id is None:
            parent_id = old_event.parent_id

        event = old_event.clone(task, parent_id)
        self.register_event(event)

        if parent_id is None:
            workflow = self.process.workflow
        else:
            workflow = self.events[parent_id].task

        def process_event(_event) -> ActiveEventState:
            # if there is no processing needed, we skip to routing
            if isinstance(event.task, StartEvent) or isinstance(event.task, EndEvent):
                return event.state.route(event.context)

            if isinstance(event.task, NonWaitingGateway):
                return event.state.route(event.context)

            # if we need to wait, we wait.
            if isinstance(event.task, WaitingGateway):
                return event.state.wait_check()

            # normally we shouldn't wait for tasks, since it's counter BPMN, so
            # we allow configuring waiting for it.
            if self.config.wait_tasks and (
                    isinstance(event.task, Task) or
                    isinstance(event.task, Workflow)
            ):
                return event.state.wait_check()

            return event.state.run()

        def get_other_task_waiting(source: ActiveEvent) -> Optional[ActiveEvent]:
            for ev in self.events.values():
                if ev == source:
                    continue

                if ev.task == source.task and ev.state.state in ACTIVE_STATES:
                    return ev

            return None

        def wait_task(_event) -> Optional[ActiveEventState]:
            # is another waiting task already present?
            other_waiting = get_other_task_waiting(event)

            potential_predecessors = list(map(
                lambda e: e.task,
                filter(
                    lambda e: e.state.state in ACTIVE_STATES and e.task != event.task,
                    self.events.values()
                )))

            if other_waiting:
                new_data = WorkflowData.merge(other_waiting.context.data, event.context.data)
                other_waiting.context.data = new_data

                event.state.done()

            # if we have predecessors, we stay in waiting
            if workflow.are_predecessors(event.task, potential_predecessors):
                return None

            if other_waiting:
                other_waiting.state.run()
                return None

            return event.state.run()

        def run_task(_event) -> None:
            if isinstance(event.task, Workflow):
                for start_task in event.task.start_tasks.values():
                    # this automatically registers our events for execution
                    self.clone_event(event, start_task, parent_id=event.id)
                return

            if isinstance(event.task, Task):
                future = WorkflowExecutor.pool.submit(
                    self.tasks_impl[event.task.id].invoke,
                    event.context)
                self.futures[future] = event.id
                event.future = future
                return

            if isinstance(event.task, UserTask):
                future = Future()
                self.futures[future] = event.id
                event.future = future

                self.ut_provider.register_event(self, event)

                return

            return event.state.route(event.context)

        def error_task(_event) -> None:
            # if we have a boundary error task, we use that one for processing.
            if event.task.error_task:
                self.clone_event(event, event.task.error_task)

                if event.task.error_task.cancel_activity:
                    event.state.done_check(None)
                else:
                    event.state.route()

                return

            if event.parent_id in self.events:
                parent_event = self.get_parent(event.id)

                # we move the parent into error
                parent_event.state.error(_event.data)

                # FIXME: not sure if this is enough.
                for potential_child in list(self.events.values()):
                    if potential_child.parent_id != event.id:
                        continue

                    potential_child.state.error(_event.data)

            event.state.done_check(None)  # we kill the current event

        def route_task(_event) -> None:
            event.context = _event.data

            if isinstance(event.task, ExclusiveGateway):
                gateway = cast(Gateway, event.task)
                outgoing_edges = GatewayController.route_single_output(
                    workflow, gateway, event)
            else:
                outgoing_edges = GatewayController.route_all_outputs(
                    workflow, event.task, event)

            for outgoing_edge in outgoing_edges:
                target_task = workflow.tasks[outgoing_edge.target_id]
                self.clone_event(event, target_task)

            event.state.done_check(outgoing_edges)

        def done_check(_event) -> None:
            outgoing_edges = _event.data

            if not outgoing_edges:
                return event.state.done_end_task()

            return event.state.done()

        def done_end_task(_event) -> None:
            # we should check all the WAITING processes if they finished.
            waiting_events = list(filter(
                lambda e: e.state.state == ActiveEventState.WAITING and e.task != event.task,
                self.events.values()
            ))

            for waiting_event in waiting_events:
                potential_predecessors = list(map(
                    lambda e: e.task,
                    filter(
                        lambda e: e.state.state in DONE_STATES and e.task != waiting_event.task,
                        self.events.values()
                    )))

                if not workflow.are_predecessors(waiting_event.task, potential_predecessors):
                    waiting_event.state.run()

            # check sub-process termination
            found = False
            for ev in self.events.values():
                if ev.parent_id == event.parent_id and ev != event:
                    found = True
                    break

            # we merge into the parent event if it's an end state.
            if parent_id is not None:
                self.events[parent_id].context.data = WorkflowData.merge(
                    self.events[parent_id].context.data,
                    event.context.data
                )

                if not found:
                    parent_event = self.events[parent_id]
                    parent_event.state.route(parent_event.context)

            event.state.done()

        def done_task(_event) -> None:
            self.unregister_event(event)

        event.state.after_enter(ActiveEventState.PROCESSING, process_event)
        event.state.after_enter(ActiveEventState.WAITING, wait_task)
        event.state.after_enter(ActiveEventState.RUNNING, run_task)
        event.state.after_enter(ActiveEventState.ERROR, error_task)
        event.state.after_enter(ActiveEventState.ROUTING, route_task)
        event.state.after_enter(ActiveEventState.DONE_CHECK, done_check)
        event.state.after_enter(ActiveEventState.DONE_END_TASK, done_end_task)
        event.state.after_enter(ActiveEventState.DONE, done_task)

        return event



from .BaseTask import BaseTask


class StartEvent(BaseTask):
    """
    StartEvent documentation.
    """
    def __init__(self,
                 _id: str,
                 name: str) -> None:
        super(StartEvent, self).__init__(_id, name)




import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file
from test.adhesive.steps.check_equals import assert_equal_steps
from .test_tasks import adhesive, _async


class TestWorkflowExecutorSubProcess(unittest.TestCase):
    def test_sub_process_execution(self):
        """
        Load a workflow that contains a sub process and execute it.
        """
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/adhesive_subprocess.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            'Ensure Docker Tooling': 1,
            'Build Germanium Image': 1,
            'Prepare Firefox': 1,
            'Test Firefox': 1,
            'Test Chrome': 1,
        }, data.steps)
        self.assertFalse(workflow_executor.events.keys())

    def test_sub_process_execution_non_wait(self):
        """
        Load a workflow that contains a sub process and execute it.
        """
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/adhesive_subprocess.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process, wait_tasks=True)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            'Ensure Docker Tooling': 1,
            'Build Germanium Image': 1,
            'Prepare Firefox': 1,
            'Test Firefox': 1,
            'Test Chrome': 1,
        }, data.steps)
        self.assertFalse(workflow_executor.events.keys())

if __name__ == '__main__':
    unittest.main()



import uuid
from typing import Optional

from adhesive.graph.BaseTask import BaseTask
from adhesive.model.ActiveEventStateMachine import ActiveEventStateMachine
from adhesive.steps.WorkflowContext import WorkflowContext


class ActiveEvent:
    """
    An event that passes through the system. It can fork
    in case there are multiple executions going down.
    """
    def __init__(self,
                 parent_id: Optional['str'],
                 task: BaseTask) -> None:
        self.id: str = str(uuid.uuid4())
        self.parent_id = parent_id

        if not isinstance(task, BaseTask):
            raise Exception(f"Not a task: {task}")

        self._task = task
        self.context = WorkflowContext(task)

        self.state = ActiveEventStateMachine()
        self.future = None

    def __getstate__(self):
        return {
            "id": self.id,
            "parent_id": self.parent_id,
            "_task": self._task,
            "context": self.context
        }

    def clone(self,
              task: BaseTask,
              parent_id: str) -> 'ActiveEvent':
        """
        Clone the current event for another task id target.
        """
        result = ActiveEvent(parent_id, task)
        result.context = self.context.clone(task)

        return result

    @property
    def task(self) -> BaseTask:
        return self._task

    @task.setter
    def task(self, task: BaseTask) -> None:
        if not isinstance(task, BaseTask):
            raise Exception(f"Not a task: {task}")

        self.context.task = task
        self._task = task

    def __repr__(self) -> str:
        return f"ActiveEvent({self.id}, {self.state.state}): {self.task}"



import re
from typing import Callable, Optional, List, Tuple, Any

from .WorkflowContext import WorkflowContext


class AdhesiveTask:
    """
    A task implementation.
    """
    def __init__(self,
                 code: Callable,
                 *expressions: str) -> None:
        self.re_expressions = list(map(re.compile, expressions))
        self.code = code

    def matches(self, name: str) -> Optional[List[str]]:
        """
        Checks if this implementation matches any of the expressions
        bounded to this task. If yes, it returns the potential variables
        extracted from the expression.
        :param name:
        :return:
        """
        for re_expression in self.re_expressions:
            m = re_expression.match(name)

            if m:
                return list(m.groups())

        return None

    def invoke(
            self,
            context: WorkflowContext) -> WorkflowContext:
        step_name = context.task.name
        params = self.matches(step_name)

        self.code(context, *params)  # type: ignore

        return context

    def invoke_user_task(
            self,
            context: WorkflowContext,
            ui: Any) -> WorkflowContext:
        step_name = context.task.name
        params = self.matches(step_name)

        self.code(context, ui, *params)  # type: ignore

        return context



from typing import TypeVar

from concurrent.futures import Future

from adhesive.graph.BaseTask import BaseTask

T = TypeVar('T')


class TaskFuture:
    """
    A Future that is bound to a task. It allows retrieving what
    task the future is assigned to. The Future should regardless
    return the ActiveEvent.

    Some tasks / gateways consume multiple elements before firing
    their event.
    """
    def __init__(self,
                 task: BaseTask,
                 future: Future):
        self.task = task
        self.future = future

    @staticmethod
    def resolved(task: BaseTask, item: T) -> 'TaskFuture':
        future = Future()
        future.set_result(item)

        return TaskFuture(task, future)

    def __str__(self) -> str:
        return f"TaskFuture {self.task}"



from typing import cast
import unittest

from adhesive.graph.BoundaryEvent import ErrorBoundaryEvent
from adhesive.graph.ExclusiveGateway import ExclusiveGateway
from adhesive.graph.UserTask import UserTask
from adhesive.graph.ParallelGateway import ParallelGateway
from adhesive.graph.SubProcess import SubProcess
from adhesive.graph.Task import Task
from adhesive.xml.bpmn import read_bpmn_file


class TestReadingBpmn(unittest.TestCase):
    """
    Test if we can read a BPMN file correctly.
    """

    def test_reading_bpmn(self) -> None:
        """
        Try to see if reading a basic BPMN works.
        """
        workflow = read_bpmn_file("test/adhesive/xml/adhesive.bpmn")

        self.assertEqual(6, len(workflow.tasks))
        self.assertEqual(6, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        first_task = workflow.tasks["_3"]
        self.assertEqual("Build Germanium Image", first_task.name)

        self.assertTrue(workflow)

    def test_reading_subprocess_bpmn(self) -> None:
        workflow = read_bpmn_file("test/adhesive/xml/adhesive_subprocess.bpmn")

        self.assertEqual(5, len(workflow.tasks))
        self.assertEqual(4, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        subprocess = cast(SubProcess, workflow.tasks["_7"])
        self.assertEqual("Test Browsers", subprocess.name)

        self.assertEqual(3, len(subprocess.tasks))
        self.assertEqual(1, len(subprocess.edges))
        self.assertEqual(2, len(subprocess.start_tasks))
        self.assertEqual(2, len(subprocess.end_events))

    def test_reading_exclusive_gateway_bpmn(self) -> None:
        workflow = read_bpmn_file("test/adhesive/xml/exclusive_gateway.bpmn")

        self.assertEqual(6, len(workflow.tasks))
        self.assertEqual(6, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        task_route = workflow.edges["_9"]
        self.assertEqual('data.route == "task"', task_route.condition)

        task_route = workflow.edges["_10"]
        self.assertEqual('', task_route.condition)

    def test_reading_parallel_gateway_bpmn(self) -> None:
        workflow = read_bpmn_file("test/adhesive/xml/gateway-parallel.bpmn")

        self.assertEqual(9, len(workflow.tasks))
        self.assertEqual(12, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        self.assertTrue(isinstance(workflow.tasks["_9"], ParallelGateway))

    def test_reading_gateway_exclusive_sign_bpmn(self) -> None:
        workflow = read_bpmn_file("test/adhesive/xml/gateway-exclusive-sign.bpmn")

        self.assertEqual(7, len(workflow.tasks))
        self.assertEqual(8, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        self.assertTrue(isinstance(workflow.tasks["_3"], ExclusiveGateway))

    def test_reading_gateway_inclusive_sign_bpmn(self) -> None:
        workflow = read_bpmn_file("test/adhesive/xml/gateway-inclusive.bpmn")

        self.assertEqual(8, len(workflow.tasks))
        self.assertEqual(10, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        self.assertTrue(isinstance(workflow.tasks["_3"], ParallelGateway))

    def test_reading_error_event_interrupting(self) -> None:
        workflow = read_bpmn_file("test/adhesive/xml/error-event-interrupting.bpmn")

        self.assertEqual(6, len(workflow.tasks))
        self.assertEqual(5, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        boundary_event: ErrorBoundaryEvent = workflow.tasks["_6"]
        self.assertTrue(isinstance(
            workflow.tasks["_6"],
            ErrorBoundaryEvent
        ))

        self.assertTrue(boundary_event.cancel_activity)
        self.assertFalse(boundary_event.parallel_multiple)

        parent_event: Task = workflow.tasks['_3']
        self.assertEqual(parent_event.error_task, boundary_event)

    def test_reading_human_task(self) -> None:
        workflow = read_bpmn_file("test/adhesive/xml/user-task.bpmn")

        self.assertEqual(3, len(workflow.tasks))
        self.assertEqual(2, len(workflow.edges))
        self.assertEqual(1, len(workflow.start_tasks))
        self.assertEqual(1, len(workflow.end_events))

        self.assertTrue(isinstance(
            workflow.tasks["_3"],
            UserTask
        ))

    def test_reading_unsupported_elements_fails(self) -> None:
        with self.assertRaises(Exception):
            read_bpmn_file("test/adhesive/xml/unsupported-call-activity.bpmn")


if __name__ == '__main__':
    unittest.main()



from .BaseTask import BaseTask


class EndEvent(BaseTask):
    """
    EndEvent
    """
    def __init__(self,
                 _id: str,
                 name: str) -> None:
        super().__init__(_id, name)




from typing import Optional


class Edge():
    """
    An edge between two tasks in a Workflow.
    """
    def __init__(self,
                 _id: str,
                 source_id: str,
                 target_id: str,
                 condition: Optional[str] = None) -> None:
        self.id = _id
        self.source_id = source_id
        self.target_id = target_id
        self.condition = condition if condition else ''




from adhesive.graph.BaseTask import BaseTask


class BoundaryEvent(BaseTask):
    def __init__(self,
                 _id: str,
                 name: str) -> None:
        super(BoundaryEvent, self).__init__(_id, name)
        self.attached_task_id = 'not attached'

        self.cancel_activity = True
        self.parallel_multiple = False


class ErrorBoundaryEvent(BoundaryEvent):
    pass




from .Workflow import Workflow


class SubProcess(Workflow):
    def __init__(self,
                 id: str,
                 name: str):
        super(SubProcess, self).__init__(id)
        self.name = name



import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file

from .test_tasks import adhesive, _async
from .check_equals import assert_equal_steps


class TestExclusiveSignGateway(unittest.TestCase):
    """
    Test if the workflow executor can process exclusive gateways.
    """
    def test_exclusive_gateway(self):
        """
        Load a workflow with a gateway and test it..
        """
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/gateway-exclusive-sign.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            "Test Chrome": 1,
            "Test Firefox": 1,
            "Build Germanium Image": 1,
        }, data.steps)
        self.assertFalse(workflow_executor.events)

    def test_exclusive_gateway_non_wait(self):
        """
        Load a workflow with a gateway and test it..
        """
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/gateway-exclusive-sign.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process, wait_tasks=False)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            "Test Chrome": 1,
            "Test Firefox": 1,
            "Build Germanium Image": 2,
        }, data.steps)
        self.assertFalse(workflow_executor.events)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Dict


def merge_dict(dict1: Dict, dict2: Dict) -> None:
    for k2, v2 in dict2.items():
        if k2 not in dict1:
            dict1[k2] = v2
            continue

        v1 = dict1[k2]

        if isinstance(v1, list) and isinstance(v2, list):
            v1.extend(v2)
            continue

        if isinstance(v1, set) and isinstance(v2, set):
            v1.update(v2)
            continue

        if isinstance(v1, dict) and isinstance(v2, dict):
            merge_dict(v1, v2)
            continue

        raise Exception(f"Unable to merge '{k2}' with different types.")


class WorkflowData:
    def __init__(self,
                 initial_data: Optional[Dict] = None):
        self._data = dict(initial_data) if initial_data else dict()

    def as_dict(self) -> Dict:
        return self._data

    def __getattr__(self, item: str):
        if item.startswith("__") or item == '_data':
            return super(WorkflowData, self).__getattribute__(item)

        if item not in self._data:
            return None

        return self._data[item]

    def __setattr__(self, key, value):
        if key == "_data":
            super(WorkflowData, self).__setattr__(key, value)
            return

        self._data[key] = value

    def __setitem__(self, key, value):
        self._data[key] = value

    def __html__(self):
        return self.as_dict()

    @staticmethod
    def merge(*data_items: 'WorkflowData') -> 'WorkflowData':
        result = WorkflowData()

        for data in data_items:
            merge_dict(result._data, data._data)

        return result



import concurrent
import time
import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file

from .test_tasks import adhesive, _async


class TestWorkflowExecutor(unittest.TestCase):
    def test_parallel_tasks(self):
        """
        Load a bunch of tasks in parallel.
        :return:
        """
        WorkflowExecutor.pool = concurrent.futures.ProcessPoolExecutor(max_workers=6)
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/parallel5.bpmn")

        start_time = time.time() * 1000.0
        _async(WorkflowExecutor(adhesive.process).execute())
        end_time = time.time() * 1000.0

        # the whole thing should be faster than 2 secs
        total_time = end_time - start_time
        self.assertTrue(total_time < 2000,
                        f"Total time ({total_time}) should be less than 2000ms.")
        self.assertTrue(total_time >= 1000,
                        f"Total time ({total_time}) should be more than 1000ms.")

    def test_parallel_sub_processes_tasks(self):
        """
        Load a bunch of tasks in parallel.
        :return:
        """
        WorkflowExecutor.pool = concurrent.futures.ProcessPoolExecutor(max_workers=6)
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/parallel5-sub-processes.bpmn")

        start_time = time.time() * 1000.0
        _async(WorkflowExecutor(adhesive.process).execute())
        end_time = time.time() * 1000.0

        # the whole thing should be faster than 2 secs
        total_time = end_time - start_time
        self.assertTrue(total_time < 2000,
                        f"Total time ({total_time}) should be less than 2000ms.")
        self.assertTrue(total_time >= 1000,
                        f"Total time ({total_time}) should be more than 1000ms.")

    def test_parallel_tasks_non_wait(self):
        """
        Load a bunch of tasks in parallel.
        :return:
        """
        WorkflowExecutor.pool = concurrent.futures.ProcessPoolExecutor(max_workers=6)
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/parallel5.bpmn")

        start_time = time.time() * 1000.0
        _async(WorkflowExecutor(adhesive.process, wait_tasks=False).execute())
        end_time = time.time() * 1000.0

        # the whole thing should be faster than 2 secs
        total_time = end_time - start_time
        self.assertTrue(total_time < 2000,
                        f"Total time ({total_time}) should be less than 2000ms.")
        self.assertTrue(total_time >= 1000,
                        f"Total time ({total_time}) should be more than 1000ms.")

    def test_parallel_sub_processes_tasks_non_wait(self):
        """
        Load a bunch of tasks in parallel.
        :return:
        """
        WorkflowExecutor.pool = concurrent.futures.ProcessPoolExecutor(max_workers=6)
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/parallel5-sub-processes.bpmn")

        start_time = time.time() * 1000.0
        _async(WorkflowExecutor(adhesive.process, wait_tasks=False).execute())
        end_time = time.time() * 1000.0

        # the whole thing should be faster than 2 secs
        total_time = end_time - start_time
        self.assertTrue(total_time < 2000,
                        f"Total time ({total_time}) should be less than 2000ms.")
        self.assertTrue(total_time >= 1000,
                        f"Total time ({total_time}) should be more than 1000ms.")


if __name__ == '__main__':
    unittest.main()



import unittest

from adhesive.model.WorkflowExecutor import WorkflowExecutor
from adhesive.xml.bpmn import read_bpmn_file
from .check_equals import assert_equal_steps
from .test_tasks import adhesive, _async


class TestWorkflowExecutorBasic(unittest.TestCase):
    def test_link_back_execution(self):
        adhesive.process.workflow = read_bpmn_file("test/adhesive/xml/link-back.bpmn")

        workflow_executor = WorkflowExecutor(adhesive.process)
        data = _async(workflow_executor.execute())

        assert_equal_steps({
            "Increment X by 1": 5,
            "Build Germanium Image": 1,
        }, data.steps)
        self.assertFalse(workflow_executor.events)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, List, Dict, Any

import npyscreen as npyscreen

from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.model.UserTaskProvider import UserTaskProvider
from adhesive.model.WorkflowExecutor import WorkflowExecutor


# FIXME: Not sure if this is the best place
# FIXME: UIBuilder should be probably a common class with the tests
from adhesive.steps.WorkflowData import WorkflowData


class UIBuilder:
    def __init__(self,
                 event: ActiveEvent):
        self.context = event.context
        self.ui_controls: Dict[str, Any] = dict()

        self.form = npyscreen.Form(name=event.task.name)

    @property
    def data(self) -> WorkflowData:
        result_dict = dict()

        for name, ui_control in self.ui_controls.items():
            result_dict[name] = ui_control.value

        return WorkflowData(result_dict)

    def add_text_input(self,
                       name: str,
                       title: Optional[str] = None,
                       value: str = '') -> None:
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        self.ui_controls[name] = self.form.add_widget(
            npyscreen.TitleText,
            name=title,
            value=value)

    def add_combo_box(self,
                      name: str,
                      title: Optional[str] = None,
                      values: Optional[List[Any]] = None):
        if not name:
            raise Exception("You need to pass a name for the input.")

        if not title:
            title = name

        self.ui_controls[name] = self.form.add_widget(
            npyscreen.TitleCombo,
            name=title,
            values=values)

    def add_default_button(self,
                           name: str) -> None:
        self.context.data[name] = name


class ConsoleUserTaskProvider(UserTaskProvider):
    def __init__(self):
        super(ConsoleUserTaskProvider, self).__init__()

    def register_event(self,
                       executor: WorkflowExecutor,
                       event: ActiveEvent) -> None:
        def run_on_curses(x):
            try:
                ui = UIBuilder(event)

                adhesive_task = executor.tasks_impl[event.task.id]
                context = adhesive_task.invoke_user_task(event.context, ui)

                # call the actual UI
                ui.form.edit()

                # put the values in the data.
                context.data = WorkflowData.merge(context.data, ui.data)

                event.future.set_result(context)
            except Exception as e:
                event.future.set_exception(e)

        npyscreen.wrapper_basic(run_on_curses)



from setuptools import setup
from setuptools import find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

packages = find_packages()

setup(
    name='adhesive',
    version='0.1.0',
    description='adhesive',
    long_description=readme,
    author='Bogdan Mustiata',
    author_email='bogdan.mustiata@gmail.com',
    license='BSD',
    entry_points={
        "console_scripts": [
            "adhesive = adhesive.mainapp:main"
        ]
    },
    install_requires=[
        "networkx==2.3",
        "npyscreen==4.10.5"],
    packages=packages,
    package_data={
        '': ['*.txt', '*.rst']
    })



import re
from typing import Tuple
from xml.etree import ElementTree

from adhesive.graph.BoundaryEvent import BoundaryEvent, ErrorBoundaryEvent
from adhesive.graph.Edge import Edge
from adhesive.graph.EndEvent import EndEvent
from adhesive.graph.ExclusiveGateway import ExclusiveGateway
from adhesive.graph.UserTask import UserTask
from adhesive.graph.ParallelGateway import ParallelGateway
from adhesive.graph.StartEvent import StartEvent
from adhesive.graph.SubProcess import SubProcess
from adhesive.graph.Task import Task
from adhesive.graph.Workflow import Workflow

TAG_NAME = re.compile(r'^(\{.+\})?(.+)$')
SPACE = re.compile(r"\s+", re.IGNORECASE)

ignored_elements = {
    # we obviously ignore extensions.
    "extensionElements",
    # we ignore the incoming and outgoing from inside the subprocesses
    # because we use the sequenceFlow elements to trace the connections.
    "incoming",
    "outgoing",
}

boundary_ignored_elements = set(ignored_elements)
boundary_ignored_elements.add("outputSet")


def read_bpmn_file(file_name: str) -> Workflow:
    """ Read a BPMN file as a build workflow. """
    root_node = ElementTree.parse(file_name).getroot()
    process = find_node(root_node, 'process')

    return read_process(process)


def find_node(parent_node, name: str):
    for node in parent_node.getchildren():
        _, node_name = parse_tag(node)
        if node_name == name:
            return node

    return None


def get_boolean(parent_node, attr: str, default_value: bool) -> bool:
    attr_value = parent_node.get(attr)

    if attr_value is None:
        return default_value

    if attr_value.upper() == "TRUE":
        return True
    elif attr_value.upper() == "FALSE":
        return False

    raise Exception(f"Not a boolean value for {attr}: {attr_value}")


def read_process(process) -> Workflow:
    node_ns, node_name = parse_tag(process)

    if "process" == node_name:
        result = Workflow(process.get('id'))
    elif "subProcess" == node_name:
        result = SubProcess(process.get('id'), normalize_name(process.get('name')))
    else:
        raise Exception(f"Unknown process node: {process.tag}")

    # we read first the nodes, then the boundary events,
    # then only the edges so they are findable
    # when adding the edges by id.
    for node in process.getchildren():
        process_node(result, node)

    for node in process.getchildren():
        process_boundary_event(result, node)

    for node in process.getchildren():
        process_edge(result, node)

    for task_id, task in result.tasks.items():
        if isinstance(task, BoundaryEvent):
            continue

        if result.has_incoming_edges(task):
            continue

        result.start_tasks[task.id] = task

    for task_id, task in result.tasks.items():
        if not result.has_outgoing_edges(task):
            result.end_events[task.id] = task

    return result


def process_node(result: Workflow,
                 node) -> None:
    node_ns, node_name = parse_tag(node)

    if "task" == node_name:
        process_node_task(result, node)
    elif "userTask" == node_name:
        process_user_task(result, node)
    elif "sequenceFlow" == node_name:
        pass
    elif "boundaryEvent" == node_name:
        pass
    elif "startEvent" == node_name:
        process_node_start_event(result, node)
    elif "endEvent" == node_name:
        process_node_end_event(result, node)
    elif "subProcess" == node_name:
        process_node_sub_process(result, node)
    elif "exclusiveGateway" == node_name:
        process_exclusive_gateway(result, node)
    elif "parallelGateway" == node_name or "inclusiveGateway" == node_name:
        process_parallel_gateway(result, node)
    elif node_name not in ignored_elements:
        raise Exception(f"Unknown process node: {node.tag}")


def process_boundary_event(result: Workflow,
                           node) -> None:
    node_ns, node_name = parse_tag(node)

    if "boundaryEvent" == node_name:
        process_boundary_task(result, node)


def process_edge(result: Workflow,
                 node) -> None:
    node_ns, node_name = parse_tag(node)

    if "sequenceFlow" == node_name:
        process_node_sequence_flow(result, node)


def process_node_task(w: Workflow, xml_node) -> None:
    """ Create a Task element from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = Task(xml_node.get("id"), node_name)
    w.add_task(task)


def process_user_task(w: Workflow, xml_node) -> None:
    """ Create a HumanTask element from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = UserTask(xml_node.get("id"), node_name)
    w.add_task(task)


def process_boundary_task(w: Workflow, xml_node) -> None:
    """ Create a Task element from the workflow """
    for node in xml_node.getchildren():
        node_ns, node_name = parse_tag(node)

        if node_name in boundary_ignored_elements:
            continue

        # node is not ignored, we either found the type
        # or we die with exception.
        task_name = normalize_name(xml_node.get("name"))

        if node_name == "errorEventDefinition":
            boundary_task = ErrorBoundaryEvent(
                xml_node.get("id"),
                task_name)

            boundary_task.attached_task_id = xml_node.get(
                "attachedToRef", default="not attached")

            boundary_task.cancel_activity = get_boolean(
                xml_node, "cancelActivity", True)
            boundary_task.parallel_multiple = get_boolean(
                xml_node, "parallelMultiple", True)

            w.add_boundary_event(boundary_task)

            return

    raise Exception("Unable to find the type of the boundary event. Only "
                    "<errorEventDefinition> is supported.")


def process_node_start_event(w: Workflow, xml_node) -> None:
    """ Create a start event from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = StartEvent(xml_node.get("id"), node_name)
    w.add_start_event(task)


def process_node_end_event(w: Workflow, xml_node) -> None:
    """ Create an end event from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = EndEvent(xml_node.get("id"), node_name)
    w.add_end_event(task)


def process_node_sub_process(w: Workflow, xml_node) -> None:
    task = read_process(xml_node)
    w.add_task(task)


def process_node_sequence_flow(w: Workflow, xml_node) -> None:
    edge = Edge(xml_node.get("id"),
                xml_node.get("sourceRef"),
                xml_node.get("targetRef"))

    condition_node = find_node(xml_node, "conditionExpression")

    if condition_node is not None:
        edge.condition = condition_node.text

    w.add_edge(edge)


def process_exclusive_gateway(w: Workflow, xml_node) -> None:
    """ Create an exclusive gateway from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = ExclusiveGateway(xml_node.get("id"), node_name)
    w.add_task(task)


def process_parallel_gateway(w: Workflow, xml_node) -> None:
    """ Create an end event from the workflow """
    node_name = normalize_name(xml_node.get("name"))
    task = ParallelGateway(xml_node.get("id"), node_name)
    w.add_task(task)


def normalize_name(name: str) -> str:
    if not name:
        return "<noname>"

    return SPACE.sub(' ', name)


def parse_tag(node) -> Tuple[str, str]:
    m = TAG_NAME.match(node.tag)

    if not m:
        raise Exception(f"Unable to parse tag name `{node}`")

    return m[1], m[2]



import adhesive
import time
import asyncio
import uuid


def _async(fn):
    data = asyncio.get_event_loop().run_until_complete(fn)
    return data


@adhesive.task(
    'Ensure Docker Tooling',
    'Test Chrome',
    'Test Firefox',
    'Build Germanium Image',
    'Prepare Firefox',
    # exclusive gateway
    'Exclusive\ Task\ Branch',
    'Populate\ task\ data',
    'Exclusive\ default\ branch',
    'Cleanup Broken Tasks'
)
def basic_task(context) -> None:
    add_current_task(context)


@adhesive.task(r'^Parallel \d+$')
def parallel_task(context) -> None:
    time.sleep(1)
    if not context.data.steps:
        context.data.steps = set()

    context.data.steps.add(context.task.name)


@adhesive.task(r'^Throw Some Exception$')
def throw_some_exception(context) -> None:
    add_current_task(context)

    raise Exception("broken")


@adhesive.task('Increment\ X\ by\ 1')
def increment_x_by_1(context):
    add_current_task(context)

    if not context.data.x:
        context.data.x = 1
        return

    context.data.x += 1


@adhesive.usertask('Read Data From User')
def read_data_from_user(context, ui) -> None:
    ui.add_text_input("branch", title="Branch")
    ui.add_combo_box("version", title="Version", values=["12.0", "12.1", "12.2", "12.3"])

    ui.add_default_button("OK")
    ui.add_default_button("Cancel")


def add_current_task(context):
    if not context.data.steps:
        context.data.steps = dict()

    if context.task.name not in context.data.steps:
        context.data.steps[context.task.name] = set()

    context.data.steps[context.task.name].add(str(uuid.uuid4()))



from typing import Optional, Dict, Any
from adhesive.steps.WorkflowData import WorkflowData
from adhesive.graph.BaseTask import BaseTask


class WorkflowContext:
    """
    A context passed to an execution of a task. It holds the information
    about:
    - data that's being populated across tasks,
    """
    def __init__(self, task: 'BaseTask',
                 data: Optional[Dict] = None) -> None:
        self.task = task
        self.data = WorkflowData(data)

    def clone(self, task: 'BaseTask') -> 'WorkflowContext':
        return WorkflowContext(task, self.data.as_dict())

    def as_mapping(self) -> Dict[str, Any]:
        return {
            "task": self.task,
            "data": self.data
        }



from adhesive.model.ActiveEvent import ActiveEvent
from adhesive.model.WorkflowExecutor import WorkflowExecutor

import abc


class UserTaskProvider:
    @abc.abstractmethod
    def register_event(self,
                       executor: WorkflowExecutor,
                       event: ActiveEvent) -> None:
        pass



from typing import Dict, Set

import unittest

test = unittest.TestCase()


def assert_equal_steps(expected: Dict[str, int],
                       actual: Dict[str, Set[str]]) -> None:
    test.assertEqual(expected.keys(), actual.keys())
    actual_dict = dict()

    for k, v in actual.items():
        actual_dict[k] = len(v)

    test.assertEqual(expected, actual_dict)
